!+
! Module: orbit_mod
!
! Authors:  Brendan Perry - Leicester University (V0.1-V2.9)
!                           NASA/GSFC (V3.0+) 
!          
! Description: Create a FITS timeseries HK file from the ODF orbit summary.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module orbit_mod

  use dal
  use oal
  use param
  use errorhandling
  use saslib

  implicit none 
  
  private 

  character(len=*), parameter :: outfilepar  = "outfile"     , &
                                 selectpar   = "selecttimes" , &
                                 startpar    = "starttime"   , &
                                 endpar      = "endtime"     , &
                                 timesteppar = "timestep"    , & 
                                 obsidpar    = "useobsid"
  logical :: selecttime

  public orbit

contains 
  
  !---------------------------------------------------------------------
  subroutine orbit
    !---------------------------------------------------------------------
    
    type(CartesianVectorType)   :: pos , vel
    type(OdfInfoType)           :: odfinfo
    type(ProposalInfoType)      :: proposalinfo
    type(DataSetT)              :: out_set
    type(TableT)                :: out_tab
    type(ColumnT)               :: time, x, y, z, vx, vy, vz, gx, gy, gz
    logical :: useobsid
    character(len=1024) :: outfile
    integer :: i , nbins
    real(kind=double) :: obsstarttime , obsendtime , starttime , endtime , &
         timestep , currtime , ranom , decnom , totaltime, &
         gse_x, gse_y, gse_z, revolut, twoDaysBefore, twoDaysAfter
    real(kind=double), dimension(:), pointer :: time_ptr , x_ptr , y_ptr , &
         z_ptr ,  vx_ptr , vy_ptr , vz_ptr, gx_ptr , gy_ptr , gz_ptr
    
    ! get all the parameters first in subroutine getparamvalues
    
    call getparamvalues(outfile , starttime , endtime , timestep, useobsid)


    call message(AppMsg,Sparse,&
         "OAL searches for RECON orbit file first, then PREDICTED.")

    write(errstr,*) "Output file is:",trim(outfile)
    call message(AppMsg,Noisy)
    
    ! start OAL calls, open ODF, get time info
    
    call OAL_odfInfo(odfinfo)

    obsstarttime = toTimeTag(odfinfo%observationStartTime)
    obsendtime = toTimeTag(odfinfo%observationEndTime)

    write(errstr,*) "Obs runs ",trim(odfinfo%observationStartTime),&
         " to ",trim(odfinfo%observationEndTime)
    call message(AppMsg,Sparse)

    ! get the real start and stop times for the observation and
    ! compare with the user-entered values 
    ! if starttime and endtime are exactly 0., default to the
    ! actual observation start and end time derived from the
    ! ODF.
    
    if (.not. selecttime) then
       starttime = obsstarttime
       endtime = obsendtime
    else
       if (starttime .ge. endtime) then 
          call error("StarttimeGEendtime", &
               "starttime >= endtime")
       end if
       twoDaysBefore = obsstarttime - 172800.
       twoDaysAfter  = obsendtime + 172800.
       write(errstr,*) "Two Days Before ",twoDaysBefore
       call message(AppMsg,Noisy)
       write(errstr,*) "Two Days After ",twoDaysAfter
       call message(AppMsg,Noisy)
       if (starttime .lt. twoDaysBefore) then 
          call warning("StartLTorbit", &
               "Starttime before interpolatable orbit data starts")
          write(errstr,*) "Defaulting starttime to ObsStart ",obsstarttime
          call message(AppMsg,Sparse)
          starttime = obsstarttime
       end if
       if (endtime .gt. twoDaysAfter) then 
          call warning("EndGTorbit", &
               "Endtime after interpolatable orbit data ends")
          write(errstr,*) "Defaulting endtime to ObsEnd ",obsendtime
          call message(AppMsg,Sparse)
          endtime = obsendtime
       end if
       if (starttime == 0) then
          write (errstr,*) "Start time = 0, Defaulting to ",&
               obsstarttime
          call warning("StartZero",trim(errstr))
          starttime = obsstarttime
       end if
       if (endtime == 0) then
          write (errstr,*) "End time = 0, Defaulting to ",&
               obsendtime
          call warning("EndZero",trim(errstr))
          endtime = obsendtime
       end if
    end if
    
    write(errstr,*) "Start time is : ",starttime
    call message(AppMsg,Sparse)
    
    write(errstr,*) "End time is   : ",endtime
    call message(AppMsg,Sparse)
    
    ! now fix the timestep. The default is 60. seconds
    
    totaltime = endtime - starttime

    if ((totaltime <= 0.)) then 
       call error("TotalTimeLEzero", &
            "endtime - starttime <= 0.")
    end if

    write(errstr,*) "Timestep is : ",timestep
    call message(AppMsg,Verbose)
   
    if (timestep <= 0. ) then 
       write (errstr,*) "Timestep parameter value (",timestep,&
            ") LE zero, defaulting to 60 seconds."
       call warning("badTimestep",trim(errstr))
       timestep = 60.
    else if (timestep > totaltime) then
       write (errstr,*) "Timestep parameter value (",timestep,&
            ") greater than time range (",totaltime,&
            "), defaulting to 60 seconds."
       call warning("badTimestep",trim(errstr))
       timestep = 60.
    end if
    
    nbins  = 1 + nint ((endtime - starttime)/timestep)

    write(errstr,*) "Total time is : ",totaltime
    call message(AppMsg,Sparse)

    ! now open and create output orbit file

    out_set = dataSet(trim(outfile),CREATE)
    out_tab = addTable(out_set,"ORBIT",nbins)
    
    time = addColumn(out_tab,"TIME",Real64,"sec",&
         comment="Time since MJDREF")
    x = addColumn(out_tab,"GEI_X",Real64,"km",&
         comment="GEI X-axis position vector")
    y = addColumn(out_tab,"GEI_Y",Real64,"km",&
         comment="GEI Y-axis position vector")
    z = addColumn(out_tab,"GEI_Z",Real64,"km",&
         comment="GEI Z-axis position vector")
    gx = addColumn(out_tab,"GSE_X",Real64,"km",&
         comment="GSE X-axis position vector")
    gy = addColumn(out_tab,"GSE_Y",Real64,"km",&
         comment="GSE Y-axis position vector")
    gz = addColumn(out_tab,"GSE_Z",Real64,"km",&
         comment="GSE Z-axis position vector")
    vx = addColumn(out_tab,"VX",Real64,"km/s",&
         comment="X-axis velocity vector")
    vy = addColumn(out_tab,"VY",Real64,"km/s",&
         comment="Y-axis velocity vector")
    vz = addColumn(out_tab,"VZ",Real64,"km/s",&
         comment="Z-axis velocity vector")
    
    time_ptr => real64Data(time)

    x_ptr => real64Data(x)
    y_ptr => real64Data(y)
    z_ptr => real64Data(z)
    gx_ptr => real64Data(gx)
    gy_ptr => real64Data(gy)
    gz_ptr => real64Data(gz)
    vx_ptr => real64Data(vx)
    vy_ptr => real64Data(vy)
    vz_ptr => real64Data(vz)
    
    ! loop through the spacecraft reconstructed orbit file by timesteps
    ! now loop through the position data by steps
    
    currtime = starttime

    write(errstr,*) "Looping over nbins= ",nbins
    call message(AppMsg,Noisy)

    do i = 0 , nbins-1

       ! timestep over endtime boundary
       ! so get last time = endtime

       if (currtime > endtime) then
          currtime = endtime 
       end if

       time_ptr(i) = currtime
! currtime is in timeTag format 
       if (OAL_getPosition(currtime, pos , vel, revolut)) then
          ! attitude is good
          x_ptr(i) = pos%x
          y_ptr(i) = pos%y
          z_ptr(i) = pos%z
          call gei2gse(time_ptr(i),x_ptr(i),y_ptr(i),z_ptr(i),&
               gse_x,gse_y,gse_z) ! convert celestial to gse
          gx_ptr(i) = gse_x
          gy_ptr(i) = gse_y
          gz_ptr(i) = gse_z
          vx_ptr(i) = vel%x
          vy_ptr(i) = vel%y
          vz_ptr(i) = vel%z

          write(errstr,'(A,F12.1,A,3(F9.1,2x))') &
               "currtime = ",currtime," pos x,y,z = ",pos%x,pos%y,pos%z
          call message(AppMsg,Noisy)

       else
          ! attitude is bad, put in value HUGE for now...
          x_ptr(i) = HUGE(1.0d0)
          y_ptr(i) = HUGE(1.0d0)
          z_ptr(i) = HUGE(1.0d0)
          gx_ptr(i) = HUGE(1.0d0)
          gy_ptr(i) = HUGE(1.0d0)
          gz_ptr(i) = HUGE(1.0d0)
          vx_ptr(i) = HUGE(1.0d0)
          vy_ptr(i) = HUGE(1.0d0)
          vz_ptr(i) = HUGE(1.0d0)
       end if
       !
       ! increment the current time by timestep
       !
       currtime = currtime + timestep
       
    end do

    ! now write a bunch of keywords, getting most of them from 
    ! the OAL_proposalInfo call
    ! now get proposal information for later keyword additions
    
    call OAL_proposalInfo(proposalinfo)
    
    ! if useobsid is selected, the keywords from the selected observation
    ! will be included (and calculated). If not selected, and an
    ! interpolated time range is given (e.g. an entire orbit not just
    ! and obsid within an orbit) only the start/stop times, instrument,
    ! revolut and ogip keywords will be added. 

    if (useobsid) then 
       ranom = proposalinfo%boresightRa
       decnom = proposalinfo%boresightDec
       write(errstr,*) "Nominal RA: ",ranom
       call message(AppMsg,Noisy)
       write(errstr,*) "Nominal Dec: ",decnom
       call message(AppMsg,Noisy)

       call message(AppMsg,Verbose,&
            "Writing keywords from proposal info to output file")


       if (odfInfo%isSlew) then
          call message(AppMsg,Noisy,"Obs mode is SLEW")
          call setAttribute(out_tab,"OBS_MODE","SLEW", &
               "Observation mode (POINTING or SLEW)")
       else
          call message(AppMsg,Noisy,"Obs mode is POINTING")
          call setAttribute(out_tab,"OBS_MODE","POINTING", &
               "Observation mode (POINTING or SLEW)") 
       endif

       call setAttribute(out_tab,"DATE-OBS",&
            trim(odfinfo%observationStartTime), &
            "Date of observation start")
       call setAttribute(out_tab,"DATE-END",trim(odfinfo%observationEndTime), &
            "Date of observation end")
       call setAttribute(out_tab,"RA_OBJ",proposalinfo%targetRa,"deg", &
            "Right Ascension of the observed object")
       call setAttribute(out_tab,"DEC_OBJ",proposalinfo%targetDec,"deg", &
            "Declination of the observed object")
       call setAttribute(out_tab,"RA_NOM",ranom,"deg", &
            "Nominal Right Ascension")
       call setAttribute(out_tab,"DEC_NOM",decnom,"deg", &
            "Nominal Declination")
    end if

    call setAttribute(out_tab,"TELESCOP","XMM","Telescope (mission) name")
    call setAttribute(out_tab,"TSTART",time_ptr(0),"s", &
         "Start of orbit segment selected")
    call setAttribute(out_tab,"TSTOP",time_ptr(nbins-1),"s", &
         "End of orbit segment selected")
    call setAttribute(out_tab,"MJDREF",50814.,"d", &
         "MJD for XMM (1998-01-01T00:00:00)")
    call setAttribute(out_tab,"REVOLUT",int(revolut),"number", &
         "Revolution (orbit) number")
    call setAttribute(out_tab,"ORIGIN","ESAC/SSC","Origin of FITS file")
    call setAttribute(out_tab,"HDUCLASS","OGIP", &
         "format conforms to OGIP/GSFC conventions")
    call setAttribute(out_tab,"HDUCLAS1","TEMPORALDATA", &
         "TEMPORAL dataset (OGIP memo 93/003)")
    call setAttribute(out_tab,"HDUCLAS2","EPHEM", &
         "Ephemeral data")
    call addComment(out_tab,&
         "This file contains an HK timeseries of XMM")
    call addComment(out_tab,&
         "positional data extracted from either the")
    call addComment(out_tab,&
         "Periodic or Reconstructed Orbit file by the OAL")
    call addComment(out_tab,&
         "in Geocentric Equatorial Inertial (GEI) coordinates.")
    call addComment(out_tab,&
         "The GEI coords are then converted to Geosynchronous")
    call addComment(out_tab,&
         "Solar Ecliptic (GSE) coordinates using a conversion") 
    call addComment(out_tab,&
         "to F90 of the Hapgood (1992) equations, including the")
    call addComment(out_tab,&
         "(1997) Corrigendum (see task ORBIT documentation)")

    call OAL_addCommonAttributes(out_set)
    call release(out_tab)
    call release(out_set)
    
  end subroutine orbit
  
    !----------------------------------------------------------------------
  subroutine getparamvalues(outfile, starttime , endtime , timestep, useobsid) 
    !----------------------------------------------------------------------
    ! subroutine to retrieve paramaters from the orbit.par file
    !
    ! outfile   : Name of the output orbit file
    ! starttime : Start time of desired interval in ODF
    ! endtime   : End time of desired interval in ODF
    ! timestep  : Time interval between bins

    character(len=*), intent(out) :: outfile
    logical, intent(out) :: useobsid
    real (kind=double), intent(out) :: starttime , endtime , timestep

    ! get in file parameter. This can be a filename (e.g. 
    ! 0001_0000000000_R1X00100ROF.FIT or it's ODF convention
    ! ODF(RGS1,RECONSTRUCTED_ORBIT)
    
    ! get output filename next
    
    outfile = trim(stringParameter(outfilepar))
    

    ! get start, end time intervals next, in w" format 
    ! only uses sec since XMM Reference Time for now...
    
    selecttime = booleanParameter(selectpar)
    
    if (selecttime) then
       starttime = realParameter(startpar)
       endtime = realParameter(endpar)
    end if

    ! get time step parameter last, in seconds...

    timestep = realParameter(timesteppar)

    useobsid = booleanParameter(obsidpar)

    return

  end subroutine getparamvalues
  !
    !----------------------------------------------------------------------
  subroutine gei2gse(time,gei_x,gei_y,gei_z,gse_x,gse_y,gse_z)
    !----------------------------------------------------------------------
    ! subroutine to convert from Geocentric Equatorial Inertial (gei)
    ! coordinates to Geocentric Solar Ecliptic (gse) 
    !
    ! converted by B.Perry from IDL code suppled by K. Kuntz (2013)
    ! mjd   : Modified Julian Day (e.g. 50814.0)
    ! gei_x,y,z : celestial x,y,z of XMM in km.
    ! gse_x,y,z : geo solar ecliptic x,y,z of XMM in km.
    ! lambda = ecliptic long of sun  (radians)
    !    eta = obliquity of ecliptic (radians)
    ! The Sun's ecliptic longitude (lambdaO) can be calculated using the 
    ! series of formulae:
    !
    !  meanAnom = 357.528 + 35999.050T0 + 0.04107H degrees 
    !  Lambda = 280.460 + 36000.772T0 + 0.04107H degrees 
    !  lambdaO = Lambda + (1.915 - 0.0048T0) sinM + 0.020 sin2M
    !
    ! where T0 is the time in Julian centuries from 12:00 UT on 1 January 2000
    ! to the midnight Universal Time (UT) preceding the time of interest and 
    ! H is the time in hours since that preceding UT midnight. Formulae 
    ! derived from the Almanac for Computers. In the intermediate formulae, 
    ! meanAnom is the Sun's mean anomaly and Lambda its mean longitude.
    !
    ! M = 357.528+35999.050*Tzero+0.04107*UT; Sun's mean anomaly
    ! L = 280.460+36000.772*Tzero+0.04107*UT; Sun's mean longitude
    !
    real (kind=double) :: ut, lambda, eta, t_0, m, test, &
         lambda_sol, a, b, c, time, t, gei_x, gei_y, gei_z, &
         gse_x, gse_y, gse_z
    real (kind=double), parameter :: Pi = 3.1415926535897931d0
    real (kind=double), parameter :: dpi180 =  Pi/180.0d0 ! dpi/180.
    
    t = (time/86400.0) + 50814.0  ! time from odf converted to mjd
    ! t = time in julian centuries since 2000 01-01 12:00 UT
    t_0 = (t-51544.5)/36525.0
    ! ut = hours since midnight of time t. 
    ut = (t-int(t)) * 24.0 
    m = 357.528+35999.050 * t_0 + 0.04107 * ut
    m = m * dpi180 
    lambda = 280.460 + 36000.772 * t_0 + 0.04107 * ut
    lambda_sol = lambda + (1.915-0.0048*t_0) * sin(m) + 0.020 * sin(2*m)
    test = int(lambda_sol / 360.)

    if(test .gt. 0) then 
       lambda_sol = lambda_sol - 360. * test
    end if

    lambda_sol = lambda_sol * dpi180
 
    eta = 23.439 - 0.013 * t_0 ! eta = obliquity
    eta = eta * dpi180

    a = gei_x
    b = gei_y * cos(eta) + gei_z * sin(eta)
    c = 0.0 - gei_y * sin(eta) + gei_z * cos(eta)
    ! km
    gse_x = a * cos(lambda_sol) + b * sin(lambda_sol)
    gse_y = 0.0 - a * sin(lambda_sol) + b * cos(lambda_sol)
    gse_z = c

    return

  end subroutine gei2gse

end module orbit_mod
